/*
 * Stowaway Shasta bluetooth Keyboard driver for LinuxZaurus
 * by Yuzo <kayama@sannet.ne.jp>
 */

#include <linux/malloc.h>
#include <linux/module.h>
#include <linux/input.h>
#if defined (CONFIG_USE_SHARP_KEYCODE)
#include <asm/sharp_keycode.h>
#endif
#include <linux/init.h>
#include <linux/serio.h>

MODULE_AUTHOR("Yuzo <kayama@sannet.ne.jp>");
MODULE_DESCRIPTION("Bluetooth Smart Keyboard driver");
MODULE_LICENSE("GPL");

static char debug = 0;
MODULE_PARM(debug, "c");

#define SERIO_SMTKBD    0x23

#if defined (CONFIG_USE_SHARP_KEYCODE)

/*
# Keycodes which ThinkOutside shasta sends.
  
#[~]   0e 16 1e 26 25 2e 36 3d 3e 46 45 4e 55 71 [BS]
#[Tab]  0d 15 1d 24 2d 2c 35 3c 43 44 4d 54 5b 5d [\]
#[Caps]  58 1c 1b 23 2b 34 33 3b 42 4b 4c 52 5a [Enter]
#[Shift]  12 1a 22 21 2a 32 31 3a 41 49 28 59 4a [/]
#[Esc]   14 02 07 10    5c   5c  13 5e 60 2f 66 [del]

Modification in Shasta's Keymapping
---------------------------------------------------------
Shasta     ->  Zaurus                     -> Keyhelper
---------------------------------------------------------
`(x0e)     ->  Cancel(x22=34)
Tab(x0d)   ->  Calender(F9,x58=88)        -> Tab
Caps(x58)  ->  Ctrl(CONTROL,x38=56)
Escape(x14)->  Home(HOME,x28=40)
Menu(x02)  ->  Menu (F11, x1d=29)
Alt(x07)   ->  Address(F10,x59=89)        -> Launcher
Fn (x10)   ->  Fn (F22, x1e=30)
---------------------------------------------------------
Up(x28)    ->  / in English Key (x64=100)
RShift(x59)->  SLKEY_UP
/(x4a)     ->  SLKEY_DELETE
AltGr(x13) ->  ZenHan(ZENHAN,F21,x46=70)
Left(x5e)  ->  ` in English Key (x56=86)
Down(x60)  ->  SLKEY_LEFT
Right(x2f) ->  SLKEY_DOWN
Del(x66)   ->  SLKEY_RIGHT
---------------------------------------------------------

Modification for English Key (/opt/QtPalmtop/etc/keycode.tbl)
---------------------------
Original_Key    Modified
---------------------------
2		x47=71
6		x48=72
7		x49=73
8		x4a=74
9		x4b=75
0		x4c=76
-		x4D=77
=		x4E=78
[		x4F=79
]		x50=80
\		x52=82
;		x53=83
'		x55=85
`		x56=86
,		x62=98
.		x63=99
/		x64=100
---------------------------
 */

/*
- Tab (0d) -> Escape
-Up (28)-> Menu
-Caps (58)-> sis擪ցj
- AltGr (13) -> Up
-Alt(07) ->\
-Menu(02)->Right

 */

static unsigned char smtkbd_keycode[128] = {
  0               , 0                , 29 /*Menu*/      , 0               ,
  0               , 0                , 0                , 89 /*Address*/  ,
  0               , 0                , 0                , 0               ,
  0               , 88 /*Calender*/  , 34 /*Cancel*/    , 0               ,

  30 /*Fn*/       , 0                , SLKEY_LSHIFT     , 70 /*ZenHan*/   ,
  SLKEY_HOME      , SLKEY_Q          , SLKEY_1          , 0               ,
  0               , 0                , SLKEY_Z          , SLKEY_S         ,
  SLKEY_A         , SLKEY_W          , 71 /* 2 */       , 0               ,

  0               , SLKEY_C          , SLKEY_X          , SLKEY_D         ,
  SLKEY_E         , SLKEY_4          , SLKEY_3          , 0               ,
  100 /* / */     , 0                , SLKEY_V          , SLKEY_F         ,
  SLKEY_T         , SLKEY_R          , SLKEY_5          , SLKEY_DOWN      ,

  0               , SLKEY_N          , SLKEY_B          , SLKEY_H         ,
  SLKEY_G         , SLKEY_Y          , 72 /* 6 */       , 0               ,
  0               , 0                , SLKEY_M          , SLKEY_J         ,
  SLKEY_U         , 73 /* 7 */       , 74 /* 8 */       , 0               ,

  0               , 98 /* , */       , SLKEY_K          , SLKEY_I         ,
  SLKEY_O         , 76 /* 0 */       , 75 /* 9 */       , 0               ,
  0               , 99 /* . */       , SLKEY_DELETE     , SLKEY_L         ,
  83 /* ; */      , SLKEY_P          , 77 /* - */       , 0               ,

  0               , 0                , 85 /* ' */       , 0               ,
  79 /* [ */      , 78 /* = */       , 0                , 0               ,
  56 /*Ctrl*/     , SLKEY_UP         , SLKEY_ENTER      , 80 /* ] */      ,
  SLKEY_SPACE     , 82 /* \ */       , 86 /* ` */       , 0               ,

  SLKEY_LEFT      , 0                , 0                , 0               ,
  0               , 0                , SLKEY_RIGHT      , 0               ,
  0               , 0                , 0                , 0               ,
  0               , 0                , 0                , 0               ,

  0               , SLKEY_BACK_SPACE , 0                , 0               ,
  0               , 0                , 0                , 0               ,
  0               , 0                , 0                , 0               ,
  0               , 0                , 0                , 0
  };

#else

/*
  Use Standard keycode defined by <linux/input.h>

      HOME   -> LEFT META
      CIRCLE -> GRAVE
      DEL    -> ESC
*/

static unsigned char smtkbd_keycode[128] = {
  0               , 0               , KEY_F11         , 0               ,
  0               , 0               , 0               , KEY_F10 /*ALT*/ ,
  0               , 0               , 0               , 0               ,
  0               , KEY_F9 /*TAB*/  , KEY_ESC /*`*/   , 0               ,

  KEY_F22         , 0               , KEY_LEFTSHIFT   , KEY_F21         ,
  KEY_F12/*Escape*/, KEY_Q          , KEY_1           , 0               ,
  0               , 0               , KEY_Z           , KEY_S           ,
  KEY_A           , KEY_W           , 71 /*KEY_2*/    , 0               ,

  0               , KEY_C           , KEY_X           , KEY_D           ,
  KEY_E           , KEY_4           , KEY_3           , 0               ,
  KEY_SLASH/*UP*/ , 0               , KEY_V           , KEY_F           ,
  KEY_T           , KEY_R           , KEY_5           , KEY_DOWN /*LEFT->DOWN*/ ,

  0               , KEY_N           , KEY_B           , KEY_H           ,
  KEY_G           , KEY_Y           , 72 /*KEY_6*/    , 0               ,
  0               , 0               , KEY_M           , KEY_J           ,
  KEY_U           , 73 /*KEY_7*/    , 74 /*KEY_8*/    , 0               ,

  0               , 98/*COMMMA*/    , KEY_K           , KEY_I           ,
  KEY_O           , 76 /*KEY_0 */   , 75 /*KEY_9 */   , 0               ,
  0               , 99/*DOT*/       , KEY_DELETE/*SLASH->DEL*/, KEY_L           ,
  83 /*SEMICOLON*/, KEY_P           , 77 /*MINUS*/    , 0               ,

  0               , 0               , 85 /*APOSTROPHE*/ , 0             ,
  79 /*[*/        , 78 /*EQUAL*/    , 0                , 0              ,
  KEY_LEFTCTRL /*Caps*/, KEY_UP /*LShift*/, KEY_ENTER  , 80 /*]*/       ,
  KEY_SPACE       , 82 /*BACKSLASH*/, 86 /* LEFT->`*/  , 0              ,

  KEY_LEFT/*DOWN->LEFT*/, 0         , 0                , 0              ,
  0               , 0               , KEY_RIGHT /*DEL->RIGHT*/, 0       ,
  0               , 0               , 0                , 0              ,
  0               , 0               , 0                , 0              ,

  0               , KEY_BACKSPACE   , 0                , 0              ,
  0               , 0               , 0                , 0              ,
  0               , 0               , 0                , 0              ,
  0               , 0               , 0                , 0
  };

#endif

struct smart_keyboard {
  unsigned char keycode[128];
  unsigned char press[32]; // 256/8
  struct input_dev dev;
  struct serio *serio;
};

void smtkbd_interrupt(struct serio *serio, unsigned char data, unsigned int flags)
{
  struct smart_keyboard *smtkbd = serio->private;
  unsigned char keycode;
  unsigned char release;
  
  if(data>0x80){
    release = 1;
    keycode = smtkbd->keycode[data-0x80];
  }else{
    release = 0;
    keycode = smtkbd->keycode[data];
  }
  
  if(keycode){
    input_report_key(&smtkbd->dev, keycode, (release)?(0):(1));
    if(release){
      smtkbd->press[keycode/8] &= ~(1<<(keycode%8));
    }else{
      smtkbd->press[keycode/8] |= (1<<(keycode%8));
    }
    if(debug) printk(KERN_INFO "Smart Keyboard [%d] : %s\n", keycode,(release)?("Release"):("Press"));
  }
}

void smtkbd_connect(struct serio *serio, struct serio_dev *dev)
{
  struct smart_keyboard *smtkbd;
  int i;

  if (serio->type == (SERIO_RS232 | SERIO_SMTKBD)){
    if((smtkbd = kmalloc(sizeof(struct smart_keyboard), GFP_KERNEL))) {
      memset(smtkbd, 0, sizeof(struct smart_keyboard));

      serio->private = smtkbd;
      if (serio_open(serio, dev)) {
	kfree(smtkbd);
	return;
      }
    
      memcpy(smtkbd->keycode, smtkbd_keycode, sizeof(smtkbd->keycode));
      for(i=0;i<128;i++){
	set_bit(smtkbd->keycode[i], smtkbd->dev.keybit);
      }
      clear_bit(0, smtkbd->dev.keybit);
      for(i=0;i<32;i++)	smtkbd->press[i] = 0;

      smtkbd->serio = serio;

      smtkbd->dev.name = "Smart Keyboard";
      smtkbd->dev.keycode = smtkbd->keycode;
      smtkbd->dev.idbus = BUS_RS232;
      smtkbd->dev.idvendor = SERIO_SMTKBD;
      smtkbd->dev.idproduct = 0x0001;
      smtkbd->dev.idversion = 0x0100;
      smtkbd->dev.evbit[0] = BIT(EV_KEY) | BIT(EV_REP);
      smtkbd->dev.private = smtkbd;
      input_register_device(&smtkbd->dev);
    }
  }
}

void smtkbd_disconnect(struct serio *serio)
{
  struct smart_keyboard *smtkbd = serio->private;
  int i,j;
  
  //Release All Key
  for(i=0;i<32;i++){
    for(j=0;j<8;j++)
      if(smtkbd->press[i]&(1<<j)) input_report_key(&smtkbd->dev, (i*8+j), 0);
    smtkbd->press[i]=0;
  }

  input_unregister_device(&smtkbd->dev);
  serio_close(serio);
  kfree(smtkbd);
}

struct serio_dev smtkbd_dev = {
  interrupt:	smtkbd_interrupt,
  connect:	smtkbd_connect,
  disconnect:	smtkbd_disconnect
};

int __init smtkbd_init(void)
{
  serio_register_device(&smtkbd_dev);
  return 0;
}

void __exit smtkbd_exit(void)
{
  serio_unregister_device(&smtkbd_dev);
}

module_init(smtkbd_init);
module_exit(smtkbd_exit);
